package fos

import scala.util.parsing.combinator.syntactical.StandardTokenParsers
import scala.util.parsing.input._
import fos.SimplyTyped._
import fos.Term._
import fos.Type._

/** This object implements a parser and runner for Project 2.
 */
object Main extends StandardTokenParsers {
  lexical.delimiters ++= List("(", ")", "\\", ".", ":", "=", "->", "{", "}", ",", "*")
  lexical.reserved   ++= List("Bool", "Nat", "true", "false", "if", "then", "else", "succ",
                              "pred", "iszero", "let", "in", "fst", "snd")

  /** t ::=          "true"
   *               | "false"
   *               | number
   *               | "succ" t
   *               | "pred" t
   *               | "iszero" t
   *               | "if" t "then" t "else" t
   *               | ident
   *               | "\" ident ":" T "." t
   *               | t t
   *               | "(" t ")"
   *               | "let" ident ":" T "=" t "in" t
   *               | "{" t "," t "}"
   *               | "fst" t
   *               | "snd" t
   */
  def term: Parser[Term] =
    positioned(
      simpleTerm ~ rep(simpleTerm)    ^^ { case t ~ ts => (t :: ts).reduceLeft[Term](App.apply) }
      | failure("illegal start of term"))

  def simpleTerm: Parser[Term] = positioned(
      "true"          ^^^ True
    | "false"         ^^^ False
    | numericLit      ^^ { case chars => lit2Num(chars.toInt) }
    | "succ" ~ term   ^? { case "succ" ~ t => Succ(t) }
    | "pred" ~ term   ^? { case "pred" ~ t => Pred(t) }
    | "iszero" ~ term ^? { case "iszero" ~ t => IsZero(t) }
    | "if" ~ term ~ "then" ~ term ~ "else" ~ term ^? {
        case "if" ~ t1 ~ "then" ~ t2 ~ "else" ~ t3 => If(t1, t2, t3)
      }
    | ident ^^ { case id => Var(id) }
    | "\\" ~ ident ~ ":" ~ typ ~ "." ~ term ^? { case "\\" ~ x ~ ":" ~ tp ~ "." ~ t => Abs(x, tp, t) }
    | "(" ~> term <~ ")"  ^^ { case t => t }
    | "let" ~ ident ~ ":" ~ typ ~ "=" ~ term ~ "in" ~ term ^? {
        case "let" ~ id ~ ":" ~ tp ~ "=" ~ t1 ~ "in" ~ t2 => App(Abs(id, tp, t2), t1)
      }
    | "{" ~ term ~ "," ~ term ~ "}" ^? { case "{" ~ t1 ~ "," ~ t2 ~ "}" => TermPair(t1, t2) }
    | "fst" ~ term                  ^? { case "fst" ~ t => First(t) }
    | "snd" ~ term                  ^? { case "snd" ~ t => Second(t) }
    | failure("illegal start of simple term"))

  /** T       ::=   "Bool"
   *              | "Nat"
   *              | T "->" T
   *              | "(" T ")"
   */
  def typ: Parser[Type] =
    positioned(
      simpleType ~ opt("->" ~> typ)   ^^ {
        case t1 ~ Some(t2) => TypeFun(t1, t2)
        case t1 ~ None => t1
      }
    | failure("illegal start of type"))

  def simpleType: Parser[Type] = positioned(
      baseType ~ opt("*" ~ simpleType) ^? {
        case bt ~ Some("*" ~ st) => TypePair(bt, st)
        case bt ~ None => bt
      }
    | failure("illegal start of simple type"))

  def baseType: Parser[Type] = positioned(
      "Bool" ^^^ TypeBool
    | "Nat"  ^^^ TypeNat
    | "(" ~> typ <~ ")" ^^ { case t => t }
  )

  def lit2Num(n: Int): Term =
    if (n == 0) Zero else Succ(lit2Num(n - 1))


  def main(args: Array[String]): Unit = {
    val stdin = new java.io.BufferedReader(new java.io.InputStreamReader(System.in))
    val tokens = new lexical.Scanner(stdin.readLine())
    phrase(term)(tokens) match {
      case Success(trees, _) =>
        try {
          println("typed: " + typeof(Nil, trees))
          for (t <- path(trees, reduce))
            println(t)
        } catch {
          case tperror: Exception => println(tperror.toString)
        }
      case e =>
        println(e)
    }
  }

}
